#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <string>
#include <algorithm>
using namespace std;

/*

setdata(byte index, long x, long y)
draw(word index, short x, short y, dword color)
draws(word index, short x, short y, float sx, floay sy, dword color)
drawr(word index, short x, short y, float rot, dword color)
drawsr(word index, short x, short y, float sx, float sy, float rot, dword color)
setdrawmode(byte mode, byte channel)
goto(word address)
randomjump(byte prob, word address)
frameend()
end()
setcount()
inccount()
countjump(byte limit, word address)
drawvh(word index, short x, short y, short sw, short sh, dword color)
drawvy(word index, short x, short y, short sx, short sy, short w, short h, dword color)
drawcombo(short left, short top, short right, short bottom, word align,  float s, dword color)
drawline(short x1, short y1, short x2, short y2, dword color)
*/

#define PIf 3.1415926536f

enum{
	GO_SetData=1,
	GO_Draw,
	GO_DrawS,
	GO_DrawR,
	GO_DrawSR,
	GO_SetDrawMode,
	GO_Goto,
	GO_RandomJump,
	GO_FrameEnd,
	GO_End,
	GO_SetCount,
	GO_IncCount,
	GO_CountJump,
	GO_DrawVH,
	GO_DrawVY,
	GO_DrawCombo,
	GO_DrawLine,
	GO_Inst_Count,
};

enum
{
    DM_Normal=0,
    DM_NoBlend,
    DM_Add,
    DM_Subtract,
    DM_Multiply,
};

int instlen[GO_Inst_Count]={1, 10, 11, 19, 15, 23, 3, 3, 4, 1, 1, 1, 1, 4, 15, 19, 19, 13};

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

int readint(const char* str)
{
    int ret;
    if(str[0]=='0' && str[1]=='x')
    {
        sscanf(str+2, "%x", &ret);
    }
    else if(str[0]=='0' && str[1]!=0)
    {
        sscanf(str+1, "%o", &ret);
    }
    else
    {
        sscanf(str, "%d", &ret);
    }
    return ret;
}

float readfloat(const char* str)
{
    float ret;
    sscanf(str, "%f", &ret);
    return ret;
}

int readmode(const char* str)
{
    if(strcmpi(str, "normal")==0)
    {
        return DM_Normal;
    }
    if(strcmpi(str, "noblend")==0)
    {
        return DM_NoBlend;
    }
    if(strcmpi(str, "add")==0)
    {
        return DM_Add;
    }
    if(strcmpi(str, "subtract")==0)
    {
        return DM_Subtract;
    }
    if(strcmpi(str, "multiply")==0)
    {
        return DM_Multiply;
    }
}

int readchannel(const char* str)
{
    int ret=0;
    char tmp[16];
    sscanf(str, "%[rgbaRGBA]", tmp);
    if(strchr(tmp, 'r') || strchr(tmp, 'R'))
    {
        ret|=1;
    }
    if(strchr(tmp, 'g') || strchr(tmp, 'G'))
    {
        ret|=2;
    }
    if(strchr(tmp, 'b') || strchr(tmp, 'B'))
    {
        ret|=4;
    }
    if(strchr(tmp, 'a') || strchr(tmp, 'A'))
    {
        ret|=8;
    }
    return ret;
}

int readalign(const char* str)
{
    int ret=0;
    if(str[0]==0)return 0;
    switch(str[0])
    {
    case 'L':
    case 'l':
        ret=1;
        break;
    case 'C':
    case 'c':
        ret=2;
        break;
    case 'R':
    case 'r':
        ret=3;
        break;
    }
    switch(str[1])
    {
    case 'T':
    case 't':
        ret|=1<<2;
        break;
    case 'C':
    case 'c':
        ret|=2<<2;
        break;
    case 'B':
    case 'b':
        ret|=3<<2;
        break;
    }
    return ret;
}

int readlabel(const char* str, vector<string>& labelname, vector<WORD>& labeladdress)
{
    vector<string>::iterator it=find(labelname.begin(), labelname.end(), str);
    if(it==labelname.end())return -1;
    return labeladdress[it-labelname.begin()];
}

void convert(const char* input, const char* output)
{
    FILE* file=fopen(input, "r");
    vector<string> labelname;
    vector<WORD> labeladdress;
    WORD saddress=0;
	int line=0;
    while(!feof(file))
    {
		line++;
        char temp[256];
        fgets(temp, 256, file);
        char token[9][17]={0,};
		int nparam=sscanf(temp, "%16s %16s %16s %16s "
            "%16s %16s %16s %16s %16s",
            token[0], token[1], token[2], token[3],
            token[4], token[5], token[6], token[7], token[8]);
		if(token[0][0]==0)
		{
		}
        else if(token[0][strlen(token[0])-1]==':')
        {
            labelname.push_back(string(token[0], strlen(token[0])-1));
            labeladdress.push_back(saddress);
        }
        else if(strcmpi(token[0], "setdata")==0)
        {
            saddress+=instlen[GO_SetData];
			if(nparam<4)printf("Line: %d SetData must have 3 parameters.\n", line);
        }
        else if(strcmpi(token[0], "draw")==0)
        {
            saddress+=instlen[GO_Draw];
			if(nparam<5)printf("Line: %d Draw must have 4 parameters.\n", line);
        }
        else if(strcmpi(token[0], "draws")==0)
        {
            saddress+=instlen[GO_DrawS];
			if(nparam<7)printf("Line: %d DrawS must have 6 parameters.\n", line);
        }
        else if(strcmpi(token[0], "drawr")==0)
        {
            saddress+=instlen[GO_DrawR];
			if(nparam<6)printf("Line: %d DrawR must have 5 parameters.\n", line);
        }
        else if(strcmpi(token[0], "drawsr")==0)
        {
            saddress+=instlen[GO_DrawSR];
			if(nparam<8)printf("Line: %d DrawSR must have 7 parameters.\n", line);
        }
        else if(strcmpi(token[0], "setdrawmode")==0)
        {
            saddress+=instlen[GO_SetDrawMode];
			if(nparam<3)printf("Line: %d SetDrawMode must have 2 parameters.\n", line);
        }
        else if(strcmpi(token[0], "goto")==0)
        {
            saddress+=instlen[GO_Goto];
			if(nparam<2)printf("Line: %d Goto must have 1 parameter.\n", line);
        }
        else if(strcmpi(token[0], "randomjump")==0)
        {
            saddress+=instlen[GO_RandomJump];
			if(nparam<3)printf("Line: %d RandomJump must have 2 parameters.\n", line);
        }
        else if(strcmpi(token[0], "frameend")==0)
        {
            saddress+=instlen[GO_FrameEnd];
        }
        else if(strcmpi(token[0], "end")==0)
        {
            saddress+=instlen[GO_End];
        }
        else if(strcmpi(token[0], "setcount")==0)
        {
            saddress+=instlen[GO_SetCount];
        }
        else if(strcmpi(token[0], "inccount")==0)
        {
            saddress+=instlen[GO_IncCount];
        }
        else if(strcmpi(token[0], "countjump")==0)
        {
            saddress+=instlen[GO_CountJump];
			if(nparam<3)printf("Line: %d CountJump must have 2 parameters.\n", line);
        }
        else if(strcmpi(token[0], "drawvh")==0)
        {
            saddress+=instlen[GO_DrawVH];
			if(nparam<7)printf("Line: %d DrawVH must have 6 parameters.\n", line);
        }
        else if(strcmpi(token[0], "drawvy")==0)
        {
            saddress+=instlen[GO_DrawVY];
			if(nparam<9)printf("Line: %d DrawVY must have 8 parameters.\n", line);
        }
        else if(strcmpi(token[0], "drawcombo")==0)
        {
            saddress+=instlen[GO_DrawCombo];
			if(nparam<8)printf("Line: %d DrawCombo must have 7 parameters.\n", line);
        }
        else if(strcmpi(token[0], "drawline")==0)
        {
            saddress+=instlen[GO_DrawLine];
			if(nparam<6)printf("Line: %d DrawLine must have 5 parameters.\n", line);
        }
        else
        {
			printf("Line: %d Unidentified instruction name, '%s'.\n", line, token[0]);
        }
    }
    rewind(file);
    FILE* out=fopen(output, "wb");
    while(!feof(file))
    {
        char temp[256];
        fgets(temp, 256, file);
        char token[9][17]={0,};
        sscanf(temp, "%16s %16s %16s %16s "
            "%16s %16s %16s %16s %16s",
            token[0], token[1], token[2], token[3],
            token[4], token[5], token[6], token[7], token[8]);
		if(token[0][0]==0)
		{
		}
        else if(token[0][strlen(token[0])-1]==':')
        {
        }
        else if(strcmpi(token[0], "setdata")==0)
        {
            BYTE inst=GO_SetData;
            BYTE index=readint(token[1]);
            long x=readint(token[2]), y=readint(token[3]);
            fwrite(&inst, 1, 1, out);
            fwrite(&index, 1, 1, out);
            fwrite(&x, 1, 4, out);
            fwrite(&y, 1, 4, out);
        }
        else if(strcmpi(token[0], "draw")==0)
        {
            BYTE inst=GO_Draw;
            short index=readint(token[1]);
            short x=readint(token[2]), y=readint(token[3]);
            DWORD color=readint(token[4]);
            fwrite(&inst, 1, 1, out);
            fwrite(&index, 1, 2, out);
            fwrite(&x, 1, 2, out);
            fwrite(&y, 1, 2, out);
            fwrite(&color, 1, 4, out);
        }
        else if(strcmpi(token[0], "draws")==0)
        {
            BYTE inst=GO_DrawS;
            short index=readint(token[1]);
            short x=readint(token[2]), y=readint(token[3]);
            float sx=readfloat(token[4]), sy=readfloat(token[5]);
            DWORD color=readint(token[6]);
            fwrite(&inst, 1, 1, out);
            fwrite(&index, 1, 2, out);
            fwrite(&x, 1, 2, out);
            fwrite(&y, 1, 2, out);
            fwrite(&sx, 1, 4, out);
            fwrite(&sy, 1, 4, out);
            fwrite(&color, 1, 4, out);
        }
        else if(strcmpi(token[0], "drawr")==0)
        {
            BYTE inst=GO_DrawR;
            short index=readint(token[1]);
            short x=readint(token[2]), y=readint(token[3]);
            float rot=readfloat(token[4])/180*PIf;
            DWORD color=readint(token[5]);
            fwrite(&inst, 1, 1, out);
            fwrite(&index, 1, 2, out);
            fwrite(&x, 1, 2, out);
            fwrite(&y, 1, 2, out);
            fwrite(&rot, 1, 4, out);
            fwrite(&color, 1, 4, out);
        }
        else if(strcmpi(token[0], "drawsr")==0)
        {
            BYTE inst=GO_DrawSR;
            short index=readint(token[1]);
            short x=readint(token[2]), y=readint(token[3]);
            float sx=readfloat(token[4]), sy=readfloat(token[5]);
            float rot=readfloat(token[6])/180*PIf;
            DWORD color=readint(token[7]);
            fwrite(&inst, 1, 1, out);
            fwrite(&index, 1, 2, out);
            fwrite(&x, 1, 2, out);
            fwrite(&y, 1, 2, out);
            fwrite(&sx, 1, 4, out);
            fwrite(&sy, 1, 4, out);
            fwrite(&rot, 1, 4, out);
            fwrite(&color, 1, 4, out);
        }
        else if(strcmpi(token[0], "setdrawmode")==0)
        {
            BYTE inst=GO_SetDrawMode;
            BYTE mode=readmode(token[1]);
            BYTE channel=readchannel(token[2]);
            fwrite(&inst, 1, 1, out);
            fwrite(&mode, 1, 1, out);
            fwrite(&channel, 1, 1, out);
        }
        else if(strcmpi(token[0], "goto")==0)
        {
            BYTE inst=GO_Goto;
            int t=readlabel(token[1], labelname, labeladdress);
            if(t<0)
            {
				printf("Line: %d Unidentified label name, '%s'.\n", line, token[1]);
            }
            WORD address=t;
            fwrite(&inst, 1, 1, out);
            fwrite(&address, 1, 2, out);
        }
        else if(strcmpi(token[0], "randomjump")==0)
        {
            BYTE inst=GO_RandomJump;
			BYTE prob=readint(token[1]);
            int t=readlabel(token[2], labelname, labeladdress);
            if(t<0)
            {
				printf("Line: %d Unidentified label name, '%s'.\n", line, token[1]);
            }
            WORD address=t;
            fwrite(&inst, 1, 1, out);
			fwrite(&prob, 1, 1, out);
            fwrite(&address, 1, 2, out);
        }
        else if(strcmpi(token[0], "frameend")==0)
        {
            BYTE inst=GO_FrameEnd;
            fwrite(&inst, 1, 1, out);
        }
        else if(strcmpi(token[0], "end")==0)
        {
            BYTE inst=GO_End;
            fwrite(&inst, 1, 1, out);
        }
        else if(strcmpi(token[0], "setcount")==0)
        {
            BYTE inst=GO_SetCount;
            fwrite(&inst, 1, 1, out);
        }
        else if(strcmpi(token[0], "inccount")==0)
        {
            BYTE inst=GO_IncCount;
            fwrite(&inst, 1, 1, out);
        }
        else if(strcmpi(token[0], "countjump")==0)
        {
            BYTE inst=GO_CountJump;
			BYTE limit=readint(token[1]);
            int t=readlabel(token[2], labelname, labeladdress);
            if(t<0)
            {
				printf("Line: %d Unidentified label name, '%s'.\n", line, token[1]);
            }
            WORD address=t;
            fwrite(&inst, 1, 1, out);
			fwrite(&limit, 1, 1, out);
            fwrite(&address, 1, 2, out);
        }
        else if(strcmpi(token[0], "drawvh")==0)
        {
            BYTE inst=GO_DrawVH;
            short index=readint(token[1]);
            short x=readint(token[2]), y=readint(token[3]);
            short sw=readint(token[4]), sh=readint(token[5]);
            DWORD color=readint(token[6]);
            fwrite(&inst, 1, 1, out);
            fwrite(&index, 1, 2, out);
            fwrite(&x, 1, 2, out);
            fwrite(&y, 1, 2, out);
            fwrite(&sw, 1, 2, out);
            fwrite(&sh, 1, 2, out);
            fwrite(&color, 1, 4, out);
        }
        else if(strcmpi(token[0], "drawvy")==0)
        {
            BYTE inst=GO_DrawVY;
            short index=readint(token[1]);
            short x=readint(token[2]), y=readint(token[3]);
            short sx=readint(token[4]), sy=readint(token[5]);
            short w=readint(token[6]), h=readint(token[7]);
            DWORD color=readint(token[8]);
            fwrite(&inst, 1, 1, out);
            fwrite(&index, 1, 2, out);
            fwrite(&x, 1, 2, out);
            fwrite(&y, 1, 2, out);
            fwrite(&sx, 1, 2, out);
            fwrite(&sy, 1, 2, out);
            fwrite(&w, 1, 2, out);
            fwrite(&h, 1, 2, out);
            fwrite(&color, 1, 4, out);
        }
        else if(strcmpi(token[0], "drawcombo")==0)
        {
            BYTE inst=GO_DrawCombo;
            short left=readint(token[1]), top=readint(token[2]);
            short right=readint(token[3]), bottom=readint(token[4]);
            WORD align=readalign(token[5]);
            float s=readfloat(token[6]);
            DWORD color=readint(token[7]);
            fwrite(&inst, 1, 1, out);
            fwrite(&left, 1, 2, out);
            fwrite(&top, 1, 2, out);
            fwrite(&right, 1, 2, out);
            fwrite(&bottom, 1, 2, out);
            fwrite(&align, 1, 2, out);
            fwrite(&s, 1, 4, out);
            fwrite(&color, 1, 4, out);
        }
        else if(strcmpi(token[0], "drawline")==0)
        {
            BYTE inst=GO_DrawLine;
            short x1=readint(token[1]), y1=readint(token[2]);
            short x2=readint(token[3]), y2=readint(token[4]);
            DWORD color=readint(token[5]);
            fwrite(&inst, 1, 1, out);
            fwrite(&x1, 1, 2, out);
            fwrite(&y1, 1, 2, out);
            fwrite(&x2, 1, 2, out);
            fwrite(&y2, 1, 2, out);
            fwrite(&color, 1, 4, out);
        }
    }
    fclose(file);
    fclose(out);
}

int main(int argc, char** argv)
{
    if(argc==1)
    {
        char input[32];
        char output[32];
        printf("Input file: ");
        scanf("%31s", input);
        getchar();
        printf("Output file: ");
        scanf("%31s", output);
        getchar();
        convert(input, output);
    }
    else if(argc==2)
    {
        char output[32];
        printf("Output file: ");
        scanf("%31s", output);
        convert(argv[1], output);
    }
    else
    {
        convert(argv[1], argv[2]);
    }
    printf("Press any key to exit.\n");
    getchar();
    return 0;
}
